"""antiSMASH worker control abstraction"""
from __future__ import annotations
from functools import wraps
from .base import BaseMapper, DataBase, async_mixin, sync_mixin


CONTROL_TIMEOUT = 300


class BaseControl(BaseMapper):
    """Dispatcher management object"""

    ATTRIBUTES = (
        'max_jobs',
        'name',
        'running',
        'running_jobs',
        'status',
        'stop_scheduled',
        'version',
    )

    INTERNAL = (
        '_db',
        '_key',
    )

    # Meh, needs to be repeated if we want to allow subclasses to have restricted attributes
    __slots__ = ATTRIBUTES + INTERNAL

    BOOL_ARGS = {
        'running',
        'stop_scheduled'
    }

    INT_ARGS = {
        'max_jobs',
        'running_jobs',
    }

    def __init__(self, db: DataBase, name: str, max_jobs: int, version: str = "unknown") -> None:
        super(BaseControl, self).__init__(db, "control:{}".format(name))
        self.name = name
        self.stop_scheduled: bool = False
        self.running: bool = True
        self.status: str = 'running'
        self.max_jobs: int = max_jobs
        self.running_jobs: int = 0
        self.version: str = version


def expiring_async_mixin(klass):
    """Override async_mixin to expire the object"""
    def commit_expire(fn):
        @wraps(fn)
        async def wrapper(self):
            ret = await fn(self)
            await self._db.expire(self._key, CONTROL_TIMEOUT)
            return ret
        return wrapper

    async def alive(self):
        return await self._db.expire(self._key, CONTROL_TIMEOUT)

    klass = async_mixin(klass)
    klass.commit = commit_expire(klass.commit)
    klass.alive = alive

    return klass


def expiring_sync_mixin(klass):
    """Override the sync_mixin to expire the object"""
    def commit_expire(fn):
        @wraps(fn)
        def wrapper(self):
            ret = fn(self)
            self._db.expire(self._key, CONTROL_TIMEOUT)
            return ret
        return wrapper

    def alive(self):
        return self._db.expire(self._key, CONTROL_TIMEOUT)

    klass = sync_mixin(klass)
    klass.commit = commit_expire(klass.commit)
    klass.alive = alive

    return klass


@expiring_async_mixin
class AsyncControl(BaseControl):
    """Control object using fetch/commit co-routines"""
    __slots__ = ()


@expiring_sync_mixin
class SyncControl(BaseControl):
    """Control object using sync fetch/commit functions"""
    __slots__ = ()



"""antiSMASH job abstraction"""
from __future__ import annotations
from datetime import datetime
import string
from typing import Any, Type, TypeVar, Union

from .base import BaseMapper, DataBase, async_mixin, sync_mixin

TJob = TypeVar("TJob", bound="BaseJob")


class BaseJob(BaseMapper):
    """An antiSMASH job as represented in the Redis DB"""
    VALID_TAXA = {'bacteria', 'fungi', 'plant'}

    PROPERTIES = (
        'genefinder',
        'hmmdetection_strictness',
        'molecule_type',
        'sideload_simple',
        'state',
        'status',
    )

    INTERNAL = (
        '_db',
        '_id',
        '_key',
        '_taxon',
        '_legacy',
    )

    ATTRIBUTES = (
        'added',
        'all_orfs',
        'asf',
        'borderpredict',
        'cassis',
        'cc_mibig',
        'cf_cdsnr',
        'cf_npfams',
        'cf_threshold',
        'clusterblast',
        'clusterfinder',
        'clusterhmmer',
        'dispatcher',
        'download',
        'email',
        'filename',
        'from_pos',
        'full_hmmer',
        'gff3',
        'inclusive',
        'ip_addr',
        'jobtype',
        'knownclusterblast',
        'last_changed',
        'minimal',
        'needs_download',
        'original_id',
        'pfam2go',
        'rre',
        'rre_minlength',
        'rre_cutoff',
        'seed',
        'sideload',
        'smcogs',
        'smcog_trees',
        'subclusterblast',
        'target_queues',
        'tfbs',
        'tigrfam',
        'to_pos',
        'trace',
        'transatpks_da',
        'tta',
    )

    # Meh, needs to be repeated if we want to allow subclasses to have restricted attributes
    __slots__ = ATTRIBUTES + INTERNAL + tuple(['_%s' % p for p in PROPERTIES])

    BOOL_ARGS = {
        'all_orfs',
        'asf',
        'borderpredict',
        'cassis',
        'full_hmmer',
        'cc_mibig',
        'clusterblast',
        'clusterfinder',
        'clusterhmmer',
        'inclusive',
        'knownclusterblast',
        'minimal',
        'needs_download',
        'pfam2go',
        'rre',
        'smcogs',
        'smcog_trees',
        'subclusterblast',
        'tfbs',
        'tigrfam',
        'transatpks_da',
        'tta',
    }

    INT_ARGS = {
        'cf_cdsnr',
        'cf_npfams',
        'from_pos',
        'rre_minlength',
        'seed',
        'to_pos',
    }

    FLOAT_ARGS = {
        'cf_threshold',
        'rre_cutoff',
    }

    DATE_ARGS = {
        'added',
        'last_changed',
    }

    LIST_ARGS = {
        'target_queues',
        'trace',
    }

    VALID_STATES = {
        'created',
        'downloading',
        'validating',
        'waiting',
        'queued',
        'running',
        'done',
        'failed',
        'removed',
    }

    STRICTNESS_LEVELS = {
        'strict',
        'relaxed',
        'loose',
        None,
    }

    SAFE_ACCESSION_CHARS = string.ascii_letters + string.digits + "._-"

    def __init__(self, db: DataBase, job_id: str) -> None:
        super(BaseJob, self).__init__(db, 'job:{}'.format(job_id))
        self._id: str = job_id

        # taxon is the first element of the ID
        self._taxon: str = self._id.split('-')[0]
        self._legacy: bool = False

        # unless this is a legacy job id
        if not self.is_valid_taxon(self._taxon) and job_id.count('-') == 4:
            self._taxon = 'bacteria'
            self._legacy = True

        # storage for properties
        self._state: str = 'created'
        self._molecule_type: str = 'nucl'
        self._genefinder: str = 'none'
        self._hmmdetection_strictness: Union[str, None] = None
        self._sideload_simple: Union[str, None] = None
        self.status: str = 'pending'
        self.original_id: Union[str, None] = None

        # Regular attributes that differ from None
        self.added: datetime = datetime.utcnow()
        self.last_changed: datetime = datetime.utcnow()

        # Initialise list type attributes to [] instead of None
        for attr in self.LIST_ARGS:
            setattr(self, attr, [])

    # Not really async, but follow the same API as the other properties
    @property
    def job_id(self) -> str:
        return self._id

    # No setter, job_id is a read-only property

    # Not really async, but follow same API as the other properties
    @property
    def taxon(self) -> str:
        return self._taxon

    # No setter, taxon is a read-only property

    @property
    def state(self) -> str:
        if self._legacy:
            state = self.status.split(' ')[0].rstrip(':')
            if state in self.VALID_STATES:
                return state
            return "created"
        return self._state

    @state.setter
    def state(self, value: Union[str, None]) -> None:
        if value is None:
            self._legacy = True
            return

        if value not in self.VALID_STATES:
            raise ValueError('Invalid state {}'.format(value))

        self._state = value
        self.changed()

    @property
    def status(self) -> str:
        return self._status

    @status.setter
    def status(self, value: str) -> None:
        self._status = value

    @property
    def molecule_type(self) -> str:
        return self._molecule_type

    @molecule_type.setter
    def molecule_type(self, value: str) -> None:
        if value is None:
            value = 'nucl'

        if value == 'nucleotide':
            value = 'nucl'

        if value not in {'nucl', 'prot'}:
            raise ValueError('Invalid molecule_type {}'.format(value))

        self._molecule_type = value

    @property
    def genefinder(self) -> str:
        return self._genefinder

    @genefinder.setter
    def genefinder(self, value: str) -> None:
        if value not in {'prodigal', 'prodigal-m', 'glimmerhmm', 'error', 'none'}:
            if self._legacy and value in {'glimmer', 'prodigal_m'}:
                if value == 'prodigal_m':
                    value = 'prodigal-m'
            else:
                raise ValueError('Invalid genefinding method {}'.format(value))
        self._genefinder = value

    # We want to migrate from "genefinder" to "genefinding" eventually
    @property
    def genefinding(self) -> str:
        return self._genefinder

    @genefinding.setter
    def genefinding(self, value) -> None:
        self.genefinder = value

    @property
    def hmmdetection_strictness(self) -> Union[str, None]:
        return self._hmmdetection_strictness

    @hmmdetection_strictness.setter
    def hmmdetection_strictness(self, value: str) -> None:
        if value not in BaseJob.STRICTNESS_LEVELS:
            raise ValueError('Invalid strictnes level {}'.format(value))

        self._hmmdetection_strictness = value

    @property
    def sideload_simple(self) -> Union[str, None]:
        return self._sideload_simple

    @sideload_simple.setter
    def sideload_simple(self, value: str) -> None:
        # validate format is correct: ACCESSION:START-END
        parts = value.split(":")
        if len(parts) != 2:
            raise ValueError("Invalid sideload_simple value {}".format(value))

        acc, coords = parts
        for ch in acc:
            if ch not in self.SAFE_ACCESSION_CHARS:
                raise ValueError("Invalid sideload_simple accession {}".format(acc))

        if acc[0] == "-":
            raise ValueError("Accession can't start with '-' for sideload_simple")

        parts = coords.split("-")
        if len(parts) != 2:
            raise ValueError("Invalid sideload_simple coordinates {}".format(coords))

        for p in parts:
            for ch in p:
                if ch not in string.digits:
                    raise ValueError("Invalid sideload_simple coordinates {}".format(coords))

        self._sideload_simple = value

    @staticmethod
    def is_valid_taxon(taxon: str) -> bool:
        """
        Check if taxon string is one of 'bacterial', 'fungal' or 'plant'
        """
        if taxon not in BaseJob.VALID_TAXA:
            return False

        return True

    def changed(self) -> None:
        """Update the job's last changed timestamp"""
        self.last_changed = datetime.utcnow()

    def to_dict(self, extra_info=False) -> dict[str, Any]:
        ret: dict[str, Any] = super(BaseJob, self).to_dict()

        if extra_info:
            ret['job_id'] = self.job_id
            ret['taxon'] = self.taxon

        return ret

    def __str__(self) -> str:
        return "Job(id: {}, state: {})".format(self._id, self.state)

    @classmethod
    def fromExisting(cls: Type[TJob], new_id: str, existing: TJob) -> TJob:
        """"Create a copy from an existing job, with a new ID

        :param new_id: New key to use
        :param existing: Existing job to copy values from
        :return: New job object
        """
        new = super(BaseJob, cls).fromExisting(new_id, existing)
        new.original_id = existing.job_id
        return new


@async_mixin
class AsyncJob(BaseJob):
    """Job using fetch/commit as co-routines"""
    __slots__ = ()


@sync_mixin
class SyncJob(BaseJob):
    """Job using sync fetch/commit functions"""
    __slots__ = ()



"""antiSMASH notice abstraction"""
from __future__ import annotations
from datetime import datetime, timedelta
from functools import wraps
from typing import TypeVar, Union

from .base import BaseMapper, DataBase, async_mixin, sync_mixin

TNotice = TypeVar("TNotice", bound="BaseNotice")


class BaseNotice(BaseMapper):
    """Notice object"""

    PROPERTIES = (
        'category',
    )

    ATTRIBUTES = (
        'added',
        'show_from',
        'show_until',
        'teaser',
        'text',
    )

    INTERNAL = (
        '_db',
        '_id',
        '_key',
    )

    # Meh, needs to be repeated if we want to allow subclasses to have restricted attributes
    __slots__ = ATTRIBUTES + INTERNAL + tuple(['_%s' % p for p in PROPERTIES])

    DATE_ARGS = {
        'show_from',
        'show_until'
    }

    VALID_CATEGORIES = {
        'error',
        'info',
        'warning',
    }

    def __init__(self, db: DataBase, notice_id: str, *, category: str = "info",
                 teaser: str = "placeholder", text: str = "placeholder",
                 show_from: Union[datetime, None] = None,
                 show_until: Union[datetime, None] = None):
        super(BaseNotice, self).__init__(db, "notice:{}".format(notice_id))
        self._id: str = notice_id
        self.category = category

        # by default, show new notices immediately
        self.show_from: datetime = show_from if show_from else datetime.utcnow()
        # by default, show notices for a week
        self.show_until: datetime = show_until if show_until else self.show_from + timedelta(days=7)

        self.teaser: str = teaser
        self.text: str = text

    @property
    def notice_id(self) -> str:
        return self._id

    @property
    def category(self) -> str:
        return self._category

    @category.setter
    def category(self, val: str) -> None:
        if val not in self.VALID_CATEGORIES:
            raise ValueError("Invalid category {!r}".format(val))
        self._category = val


def expiring_async_mixin(klass):
    """Override async_mixin to expire the object"""
    def commit_expire(fn):
        @wraps(fn)
        async def wrapper(self):
            ret = await fn(self)
            # aio-redis expireat can't deal with datetime objects, but converting datetimes to timestamps sucks,
            # so use expire on the difference instead
            td = self.show_until - datetime.utcnow()
            await self._db.expire(self._key, int(td.total_seconds()))
            return ret
        return wrapper

    klass = async_mixin(klass)
    klass.commit = commit_expire(klass.commit)

    return klass


def expiring_sync_mixin(klass):
    """Override the sync_mixin to expire the object"""
    def commit_expire(fn):
        @wraps(fn)
        def wrapper(self):
            ret = fn(self)
            # regular redis doesn't deal with datetime objects anymore either, so also use expire on the difference
            td = self.show_until - datetime.utcnow()
            self._db.expire(self._key, int(td.total_seconds()))
            return ret
        return wrapper

    klass = sync_mixin(klass)
    klass.commit = commit_expire(klass.commit)

    return klass


@expiring_async_mixin
class AsyncNotice(BaseNotice):
    """Notice object using fetch/commit co-routines"""
    __slots__ = ()


@expiring_sync_mixin
class SyncNotice(BaseNotice):
    """Notice object using sync fetch/commit functions"""
    __slots__ = ()



"""Base Redis<->Python object mapper"""
from __future__ import annotations
from datetime import datetime
import json
from typing import Any, Type, TypeVar, Union

from redis import Redis as SyncRedis
from redis.asyncio import Redis as AsyncRedis

DataBase = Union[SyncRedis, AsyncRedis]
TMapper = TypeVar("TMapper", bound="BaseMapper")


class BaseMapper:
    """Base object mapper class"""

    ATTRIBUTES: tuple[str, ...] = ()
    PROPERTIES: tuple[str, ...] = ()
    INTERNAL: tuple[str, ...] = ('_db', '_key')

    __slots__: tuple[str, ...] = ATTRIBUTES + INTERNAL + tuple(['_%s' % p for p in PROPERTIES])

    BOOL_ARGS: set[str] = set()
    INT_ARGS: set[str] = set()
    FLOAT_ARGS: set[str] = set()
    DATE_ARGS: set[str] = set()
    LIST_ARGS: set[str] = set()

    def __init__(self, db: DataBase, key: str) -> None:
        self._db: DataBase = db
        self._key: str = key

        for attribute in self.ATTRIBUTES:
            setattr(self, attribute, None)

    def to_dict(self) -> dict[str, Any]:
        ret: dict[str, Any] = {}

        args: tuple[str, ...] = self.PROPERTIES + self.ATTRIBUTES

        for arg in args:
            if getattr(self, arg) is not None:
                arg_val = getattr(self, arg)

                # redis can't handle bool or datetime types, int and float are fine
                if arg in self.BOOL_ARGS:
                    arg_val = str(arg_val)
                elif arg in self.DATE_ARGS:
                    arg_val = arg_val.strftime("%Y-%m-%d %H:%M:%S.%f")
                elif arg in self.LIST_ARGS:
                    arg_val = json.dumps(arg_val)

                ret[arg] = arg_val

        return ret

    def _parse(self, args, values) -> None:
        for i, arg in enumerate(args):
            val = values[i]

            # Don't touch values that are unset
            if val is None and getattr(self, arg) is None:
                continue

            if val is None:
                # avoid type conversion for None values
                # this allows 'unsetting' values that used to be set
                # only create an empty list for list args
                if arg in self.LIST_ARGS:
                    val = []
            elif arg in self.BOOL_ARGS:
                val = (val != 'False')
            elif arg in self.INT_ARGS:
                val = int(val)
            elif arg in self.FLOAT_ARGS:
                val = float(val)
            elif arg in self.DATE_ARGS:
                # We're not totally fixated on sub-second resolution
                try:
                    val = datetime.strptime(val, "%Y-%m-%d %H:%M:%S.%f")
                except ValueError:
                    val = datetime.strptime(val, "%Y-%m-%d %H:%M:%S")
            elif arg in self.LIST_ARGS:
                val = json.loads(val)

            setattr(self, arg, val)

    @classmethod
    def fromExisting(cls: Type[TMapper], new_id: str, existing: TMapper) -> TMapper:
        """"Create a copy from an existing object, with a new ID

        :param new_id: New key to use
        :param existing: Existing object to copy values from
        :return: New mapper object
        """
        if not isinstance(existing, cls):
            raise ValueError("Can't copy from type {!r}".format(existing.__class__))
        args, values = zip(*existing.to_dict().items())
        new = cls(existing._db, new_id)
        new._parse(args, values)
        return new


def async_mixin(klass):
    """Mixin for using aioredis for database connectivity"""
    async def fetch(self):
        args = self.PROPERTIES + self.ATTRIBUTES

        exists = await self._db.exists(self._key)
        if exists == 0:
            raise ValueError("No {} with ID {} in database, can't fetch".
                             format(self.__class__.__name__, self._key))

        values = await self._db.hmget(self._key, *args)

        self._parse(args, values)
        return self

    async def commit(self):
        return await self._db.hset(self._key, mapping=self.to_dict())

    async def delete(self):
        return await self._db.delete(self._key)

    klass.fetch = fetch
    klass.commit = commit
    klass.delete = delete

    return klass


def sync_mixin(klass):
    """Mixin for using redis for database connectivity"""
    def fetch(self):
        args = self.PROPERTIES + self.ATTRIBUTES

        exists = self._db.exists(self._key)
        if exists == 0:
            raise ValueError("No {} with ID {} in database, can't fetch".
                             format(self.__class__.__name__, self._key))

        values = self._db.hmget(self._key, *args)

        self._parse(args, values)
        return self

    def commit(self):
        return self._db.hset(self._key, mapping=self.to_dict())

    def delete(self):
        return self._db.delete(self._key)

    klass.fetch = fetch
    klass.commit = commit
    klass.delete = delete

    return klass



__version__ = '0.1.24'

from .control import AsyncControl, SyncControl
from .job import AsyncJob, SyncJob
from .notice import AsyncNotice, SyncNotice

__all__ = [
    'AsyncControl',
    'SyncControl',
    'AsyncJob',
    'SyncJob',
    'AsyncNotice',
    'SyncNotice',
]
